/*! \file generators_b.c
    \brief Implementation der in generators_b.h definierten Zufallszahlgeneratoren
 */
/* Generator  RAN0    
   Quelle: Numerical Recipes in C 2nd edition  (Press et al 1992)  S. 278f
           Autoren: Park and Miller, Communications of the ACM 31, 1192 (1988).

   Anmerkung Sormann: dieser Generator wird in den Numerical Recipes
                      als "Minimal Standard Generator" bezeichnet.

                      Periode = 2**31-2 = 2147483646

                      SEED = jeder INTEGER-Wert ausser  123459876

                      Probleme: der Generator ist korrelationsanfaellig,
                      wenn die Zahl der random numbers > 10**7 wird.
*/   

#include "generators_b.h"

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define MASK 123459876

float ran0(long *idum)
// First CALL: IDUM can be any INTEGER except 123459876
{
	long k;
	float ans;

	*idum ^= MASK;
	k=(*idum)/IQ;
	*idum=IA*(*idum-k*IQ)-IR*k;
	if (*idum < 0) *idum += IM;
	ans=AM*(*idum);
	*idum ^= MASK;
	return ans;
}
#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef MASK
/* (C) Copr. 1986-92 Numerical Recipes Software =$j?S2#:9. */






/* Generator  RAN1    
   Quelle: Numerical Recipes in C 2nd edition  (Press et al 1992)  S. 280f
           Kombination des Park-Miller-Generators mit dem Bays-Durham Shuffle.

   Anmerkung Sormann: Laut Aussage von Press et al besteht dieser Generator 
                      sehr viele statistische Tests, solange die Zahl der
                      berechneten random numbers 10**8 nicht uebersteigt.

                      Periode = ca. 2 . 10**9

                      SEED = jeder NEGATIVE INTEGER-Wert 

*/   
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

float ran1(long *idum)
// First CALL: IDUM must be a NEGATIVE INTEGER
{
	int j;
	long k;
	static long iy=0;
	static long iv[NTAB];
	float temp;

	if (*idum <= 0 || !iy) {
		if (-(*idum) < 1) *idum=1;
		else *idum = -(*idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(*idum)/IQ;
			*idum=IA*(*idum-k*IQ)-IR*k;
			if (*idum < 0) *idum += IM;
			if (j < NTAB) iv[j] = *idum;
		}
		iy=iv[0];
	}
	k=(*idum)/IQ;
	*idum=IA*(*idum-k*IQ)-IR*k;
	if (*idum < 0) *idum += IM;
	j=iy/NDIV;
	iy=iv[j];
	iv[j] = *idum;
	if ((temp=AM*iy) > RNMX) return RNMX;
	else return temp;
}
#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
/* (C) Copr. 1986-92 Numerical Recipes Software =$j?S2#:9. */






/* Generator  RAN2    
   Quelle: Numerical Recipes in C 2nd edition  (Press et al 1992)  S. 282f
           Generators von L'Ecuyer mit dem Bays-Durham Shuffle.
           (L'Ecuyer, Communications of the ACM 31, 742 (1988)).

   Anmerkung Sormann: Press et al sind von diesem Generator so ueberzeugt,
                      dass sie demjenigen, der mittels eines statistischen
                      Tests ein Fehlverhalten des Generators nachweist,
                      1000 US Dollar bezahlen. 

                      Periode = ca. 2.3 . 10**18   (!!)

                      SEED = jeder NEGATIVE INTEGER-Wert 

*/   
// RAN2.CC  aus den Numerical Recipes C:
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

float ran2(long *idum)
// First CALL: IDUM must be NEGATIVE INTEGER
{
	int j;
	long k;
	static long idum2=123456789;
	static long iy=0;
	static long iv[NTAB];
	float temp;

	if (*idum <= 0) {
		if (-(*idum) < 1) *idum=1;
		else *idum = -(*idum);
		idum2=(*idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(*idum)/IQ1;
			*idum=IA1*(*idum-k*IQ1)-k*IR1;
			if (*idum < 0) *idum += IM1;
			if (j < NTAB) iv[j] = *idum;
		}
		iy=iv[0];
	}
	k=(*idum)/IQ1;
	*idum=IA1*(*idum-k*IQ1)-k*IR1;
	if (*idum < 0) *idum += IM1;
	k=idum2/IQ2;
	idum2=IA2*(idum2-k*IQ2)-k*IR2;
	if (idum2 < 0) idum2 += IM2;
	j=iy/NDIV;
	iy=iv[j]-idum2;
	iv[j] = *idum;
	if (iy < 1) iy += IMM1;
	if ((temp=AM*iy) > RNMX) return RNMX;
	else return temp;
}
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
/* (C) Copr. 1986-92 Numerical Recipes Software +. */





/* Generator RANBH:
   Random number generator fuer [0,1] nach Binder-Heermann, 
   "Monte-Carlo Simulation in Statistical Physics", Springer 1988, S.76

   At first CALL, IBM must be an ODD INTEGER !!
*/
float ranbh(long *ibm)
{
long mult,modulo;
float rmodul;

mult=1277;
modulo=131072;
rmodul=modulo;
(*ibm)*=mult;
(*ibm)=(*ibm) % modulo;
return (*ibm/rmodul);
}




/*  Anmerkung H. Sormann: Dieses SHUFFLE-Programm stellt eine C-Version jenes
                          F77-Programms dar, welche Sie in den "Numerical Recipes"
                          (Press et al, Cambridge Uni Press 1986) auf S. 195 finden.

                          In der vorliegenden Version arbeitet dieses Programm
                          in Kombination mit dem "Binder-Heermann-Generator" !!
*/                          
float gly,glv[98];

float shuffle(long *idum)
{
int j,pause;
float dum;

if(*idum < 0) {
  *idum=abs(*idum);
  for(j=1;j<=97;j++) dum=ranbh(idum);
  for(j=1;j<=97;j++) glv[j]=ranbh(idum);
  gly=ranbh(idum);
}
j=1 + (int)(97.0*gly);
if(j > 97 || j < 1) {
  printf("Pause in routine SHUFFLE\n");
  scanf("%d", &pause);
}
gly=glv[j];
glv[j]=ranbh(idum);
return gly;
}






// H. Sormann 15-3-99:  Variation der C++Version fuer den Marsaglia-Zaman-RNG
//                      von Herrn Danilo Neuber auf der ALPHA2:
//
//---mzran.c------------------------------------------------------------------
//
// Zufallszahlengenerator von G. Marsaglia und A. Zaman
// (siehe "Computers in Physics *,117 (1994)")


//   Periode = 2**125 = (approx) 10**37 (!!!!)

//
// Danilo R. Neuber (DRN)
//

//---Variablen fuer mzran13()---
// Die Definition als `static' bewirkt, dass die Variablen nur fuer die in
// `mzran.c' definierten Funktionen sichtbar sind.
//
typedef unsigned int  uint;

static uint      x =  521288629;
static uint      y =  362436069;
static uint      z =   16163801;
static uint      c =          1;
static uint      n = 1131199209;


//---mzran13()----------------------------------------------------------------
// Der Zufallszahlengenerator: Rueckgabewert ist eine `Zufallszahl' im Bereich
// von 0 bis 2^32-1. Der Datentyp fuer X,Y,Z,C,N  muss die richtige Groesse haben:
//                        sizeof(...) == 4  !
//
uint      mzran13()
{
    long s;
    if (y > x+c) {
	s = y-x-c;
	c = 0;
    } else {
	s = y-x-c-18;
	c = 1;
    }
    x = y;
    y = z;
    z = s;
    n = 69069*n + 1013904243;
    return z+n;
}

//---ran13set()---------------------------------------------------------------
// Setzen der Anfangswerte fuer die vier, von mzran13() verwendeten Variablen.
//
void ran13set(uint xx, uint yy, uint zz, int nn)
{
    x = xx;
    y = yy;
    z = zz;
    n = nn;
    c = y > z;
}

//---mzruni()-----------------------------------------------------------------
// Rueckgabe einer `Zufallsfliesskommazahl' im Bereich von 0.0 bis 1.0;
// basiert auf mzran13()
//
float mzruni()
{
    return 0.2328306e-9*mzran13();
}
